
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- map_body object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local map_body = {}
map_body.table = MAP_BODY
map_body.debug = true
map_body.initialized = false
map_body.parent = nil
map_body.iscurrent = true
map_body.mpoints = nil
map_body.next_points = nil
map_body.old_points = nil
map_body.old_origin = nil

map_body.points = nil
map_body.level = level

-- init
map_body.initial_maps = nil
map_body.check_initial_maps_loaded = false

-- collision
map_body.collided = false
map_body.cnormal = nil
map_body.cpoint = nil
map_body.c_mpoint = nil

local map_body_mt = { __index = map_body }
function map_body:new(level, points)
  local mbody = setmetatable({}, map_body_mt)
  mbody.points = points
  mbody.level = level
  
  
  -- check if all maps loaded
  local loaded = true
  local level_map = level:get_level_map()
  local initial_tmaps = {}
  local tmap = level_map:get_tile_map_at_position(points[1])
  initial_tmaps[1] = tmap
  for i=2,#points do
    local p = points[i]
    local tmap = level_map:get_tile_map_at_position(p, tmap)
    initial_tmaps[i] = tmap
    
    if not tmap:is_loaded() then
      loaded = false
    end
  end
  
  if not tmap:is_loaded() then
    loaded = false
  end
  
  if not loaded then
    for i=1,#initial_tmaps do
      initial_tmaps[i]:add_to_init_queue(mbody)
    end
    
    self.initial_maps = initial_tmaps
    self.check_initial_maps_loaded = true
    return mbody
  end
  
  
  mbody.init(mbody)
  
  return mbody
end

function map_body:init(check_load)
  if self.initialized then
    return
  end

  if self.check_initial_maps_loaded then
    local init_maps = self.initial_maps
    for i=1,#init_maps do
      if not init_maps[i]:is_loaded() then
        return
      end
    end
  end

  -- create map points from table of points (vector2's)
	local avg_point = vector2:new(0,0)
	local points = self.points
	local level = self.level
	local mpoints = {}
	for i=1,#points do
		local p = points[i]
		local mpoint = map_point:new(level, p)
		mpoints[#mpoints+1] = mpoint
		
		avg_point = avg_point + p
	end
	
	local origin = avg_point / #points
                
	self.mpoints = mpoints
	self.origin = origin
	
	self.initialized = true
end

function map_body:set_parent(parent)
	self.parent = parent
end

function map_body:set_origin(pos)
	self.origin = pos
end

function map_body:translate(offset)
	if self.iscurrent then
		self.old_origin = self.origin
	end

	local next_points = self.next_points
	local mpoints = self.mpoints
	
	if next_points == nil then
		next_points = {}
		for i=1,#self.mpoints do
			next_points[i] = mpoints[i]:get_position() + offset
		end
	else
		for i=1,#next_points do
			next_points[i] = next_points[i] + offset
		end
	end
	self.next_points = next_points
	self.origin = self.origin + offset
	self.iscurrent = false
end

function map_body:set_position(new_pos)
	local offset = new_pos - self.origin
	self:translate(offset)
end
function map_body:get_position()
	return self.origin
end

function map_body:update_position(new_pos)
	local old_origin = self.old_origin or self.origin

	local offset = new_pos - self.origin
	local mpoints = self.mpoints
	local is_colliding = false
	for i=1,#mpoints do
		local m = mpoints[i]
		local new = m:get_position() + offset
		m:update_position(new)
		
		local tile = m:get_tile()
		if not tile.walkable then
			is_colliding = true
		end
	end
	
	self.origin = new_pos
	
	-- move to old position if new position results in collision
	-- this can happen when more then one map point is colliding at the same time
	if is_colliding then
		local offset = old_origin - self.origin
		for i=1,#mpoints do
			local m = mpoints[i]
			local new = m:get_position() + offset
			mpoints[i]:update_position(new)
		end
		
		self.origin = old_origin
	end
	
	return not is_colliding, self.origin
end

function map_body:get_collision_data()
	if self.collided then
		return true, self.cnormal, self.cpoint, self.coffset, self.ctiles
	else
		return false
	end
end

-- true if all tiles walkable, false otherwise
function map_body:check_tiles()
	local result = true
	local mpoints = self.mpoints
	for i=1,#mpoints do
		local tile = mpoints[i]:get_tile()
		if not tile.type.walkable then
			result = false
			break
		end
	end
	
	return result
end

------------------------------------------------------------------------------
function map_body:update(dt)
  if not self.initialized then
    return
  end

	if self.iscurrent then return end
	
	-- move mpoints to new positions
	self.collided = false
	local mpoints = self.mpoints
	local next_points = self.next_points
	self.ctiles = {}
	for i=1,#mpoints do
		mpoint = mpoints[i]
		mpoint:set_position(next_points[i])
		mpoint:update(dt)
		local c, n, p, poffset, tile = mpoint:get_collision_data()
		if c then
			self.collided = true
			self.cnormal = n
			self.cpoint = p
			self.c_mpoint = mpoint
			self.coffset = poffset
			
			if not tile.walkable then
				self.ctiles[#self.ctiles+1] = tile
			end
		end
	end
	
	self.next_points = nil
	self.iscurrent = true
end

------------------------------------------------------------------------------
function map_body:draw()
  if not self.initialized then
    return
  end

	if self.debug then
		-- draw points
		lg.setColor(0,255,0,255)
		lg.setPointSize(4)
		for i=1,#self.mpoints do
			local p = self.mpoints[i]
			local pos = p:get_position()
			lg.point(pos.x, pos.y)
		end
		
		-- origin
		lg.setColor(255, 0, 0, 255)
		lg.point(self.origin.x, self.origin.y)
		
	end
	
end

return map_body























