                                    
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- sonar_gun object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local sonar_gun = {}
sonar_gun.table = 'sonar_gun'

local sg = sonar_gun
sg.position = nil
sg.direction = nil
sg.master_timer = nil
sg.ray = nil
sg.ray_data = nil

sg.num_rays = 200
sg.angle_width = 0.4
sg.speed = 0.04
sg.time = 0
sg.depth = 3
sg.sonar_speed = 2800
sg.flash_time = 1.2
sg.is_shooting = false
sg.shoot_position = nil
sg.shoot_direction = nil

sg.points = nil
sg.active_points = nil

local sonar_gun_mt = { __index = sonar_gun }
function sonar_gun:new(level, position, direction)
  local sg = setmetatable({}, sonar_gun_mt)
  
  sg.position = vector2:new(position.x, position.y)
  sg.direction = vector2:new(direction.x, direction.y)
  sg.shoot_position = vector2:new(0, 0)
  sg.shoot_direction = vector2:new(0, 0)
  sg.ray = ray:new(level, position, direction)
  
  local master_timer = level:get_master_timer()
  local ray_data = {}
  for i=1,sonar_gun.num_rays do
    local points = {}
  
    ray_data[i] = {phase = math.random() * 2 * math.pi,
                   speed = sonar_gun.speed,
                   div = sonar_gun.angle_width,
                   dir = vector2:new(0, 0),
                   points = points,
                   ray = ray:new(level)}
  end
  sg.ray_data = ray_data
  
  sg.level = level
  sg.master_timer = master_timer
  sg.timer = timer:new(master_timer, 0)
  sg.timer:start()
  
  -- color
  sg.gradient = require("flash")
  local curve_points = require("flash_curve")
  sg.flash_curve = curve:new(curve_points, 200)
  
  return sg
end

function sonar_gun:set_position(pos)
  self.position:clone(pos)
end

function sonar_gun:set_direction(dir)
  self.direction:clone(dir)
end

function sonar_gun:shoot()
  local pos = self.position
  local rd = self.ray_data
  local ray = self.ray
  local depth = self.depth
  local angle = math.atan2(self.direction.x, self.direction.y)
  local time = self.time
  local speed = self.sonar_speed
  local master_timer = self.master_timer
  local flash_time = self.flash_time
  local cx, cy, cw, ch = self.level:get_camera():get_viewport()
  local pad = 200
  
  local points = {}
  
  for i=1,#rd do
    local data = rd[i]
    local min = angle - data.div
    local max = angle + data.div
    local angle = min + (0.5 * math.sin(time * data.speed + data.phase) + 0.5) * (max - min)
    data.dir.x = math.sin(angle)
    data.dir.y = math.cos(angle)
    
    data.ray:set(pos, data.dir)
    data.ray:cast(depth)
    local line = data.ray:get_line_points()
    local ldepth = (#line / 2) - 1
    
    local dist = 0
    for i=1,ldepth do
      local b = 1 + 2 * (i - 1)
      local x1 = line[b]
      local y1 = line[b+1]
      local x2 = line[b+2]
      local y2 = line[b+3]
      
      if x2 > cx - pad and x2 < cx + cw + pad and y2 > cy - pad and y2 < cy + ch + pad then
        dist = dist + math.sqrt((x2-x1)^2 + (y2-y1)^2)
        local delay = dist / math.random(0.95 * speed, speed)
        
        local p = {}
        p.x = x2
        p.y = y2
      
        p.depth = i
        p.delay = delay
        p.timer = timer:new(master_timer, flash_time)
        p.ray_data = data
        
        points[#points + 1] = p
      end
    end
    
    
  end
  
  self.shoot_position:clone(self.position)
  self.shoot_direction:clone(self.direction)
  self.points = points
  self.active_points = {}
  self.is_shooting = true
  self.timer:start()
end

------------------------------------------------------------------------------
function sonar_gun:update(dt)
  self.time = self.time + dt
  
  if not self.is_shooting then
    return
  end
  
  -- find newly active points
  local time = self.timer:time_elapsed()
  local points = self.points
  local active = self.active_points
  
  for i=#points,1,-1 do
    local p = points[i]
    
    if p.delay <= time then
      table.remove(points, i)
      active[#active + 1] = p
      p.timer:start()
    end
  end
  
  -- update position of active points
  local rays = {}
  for i=1,#active do
    rays[active[i].ray_data] = active[i].ray_data
  end
  
  local pos = self.shoot_position
  local angle = math.atan2(self.shoot_direction.x, self.shoot_direction.y)
  local ray = self.ray
  local time = self.time
  for _,data in pairs(rays) do
    local min = angle - data.div
    local max = angle + data.div
    local angle = min + (0.5 * math.sin(time * data.speed + data.phase) + 0.5) * (max - min)
    data.dir.x = math.sin(angle)
    data.dir.y = math.cos(angle)
    
    data.ray:set(pos, data.dir)
    data.ray:cast(self.depth)
    local line = data.ray:get_line_points()
    local ldepth = (#line / 2) - 1
    local points = data.points
    for i=1,#points do
      points[i] = nil
    end
    
    for i=1,ldepth do
      local b = 1 + 2 * (i - 1)
      local x2 = line[b+2]
      local y2 = line[b+3]
      points[#points + 1] = x2
      points[#points + 1] = y2
    end
  end
  
  
  -- update active points
  local flash_curve = self.flash_curve
  local gradient = self.gradient
  for i=#active,1,-1 do
    local p = active[i]
    
    local depth = p.depth
    local points = p.ray_data.points
    local idx = 1 + 2*(depth-1)
    if points[idx] and math.random() > 0.9 then
      p.x = points[idx]
      p.y = points[idx + 1]
    end
    
    local progress = p.timer:progress()
    p.progress = flash_curve:get(progress)

    local idx = math.floor((1 - p.progress) * #gradient)
    if idx == 0 then idx = 1 end
    local color = gradient[idx]
    p.color = color
    
    if progress == 1 then
      table.remove(active, i)
    end
  end
  
  if #points == 0 and #active == 0 then
    self.is_shooting = false
  end
  
end

------------------------------------------------------------------------------
function sonar_gun:draw()
  if not self.is_shooting then
    return
  end

  local points = self.active_points
  local alpha = 60
  local dec = 15
  
  lg.setColor(200, 255, 230, 60)
  
  for i=1,#points do
    local p = points[i]
    lg.setColor(p.color)
    lg.setPointSize(self.depth - p.depth + 1)
    lg.point(p.x, p.y)
  end
  
end

return sonar_gun










